// -*-c++-*-

/*!
  \file strategy.h
  \brief team strategh Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA
 Modified Copyright (C) Hossein Mobasher

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef AGENT2D_STRATEGY_H
#define AGENT2D_STRATEGY_H

#include <rcsc/geom/vector_2d.h>

#include <boost/shared_ptr.hpp>
#include <map>
#include <vector>
#include <string>


enum PositionType {
    Position_Left = -1,
    Position_Center = 0,
    Position_Right = 1,
};

enum StrategyType {
    FastPassShoot_Strategy, // Brainstormers type
    ManMark_Strategy, // Brainstormers,AmoiensisNQ type
    Normal_Strategy,
    No_Strategy,
};


namespace rcsc {
class WorldModel;
class Formation;
}


class SoccerRole;


class Strategy {
public:
    static const std::string BEFORE_KICK_OFF_CONF;
    static const std::string GOAL_KICK_OPP_CONF;
    static const std::string GOAL_KICK_OUR_CONF;
    static const std::string GOALIE_CATCH_OPP_CONF;
    static const std::string GOALIE_CATCH_OUR_CONF;

    static const std::string NORMAL_FORMATION_CONF;
    static const std::string OFFENSE_FORMATION_CONF;
    static const std::string DEFENSE_FORMATION_CONF;
    static const std::string KICKIN_OUR_FORMATION_CONF;
    static const std::string SETPLAY_OUR_FORMATION_CONF;
private:
    typedef SoccerRole * (*RoleCreator)( boost::shared_ptr< const rcsc::Formation > );
    typedef rcsc::Formation * (*FormationCreator)();

    typedef boost::shared_ptr< rcsc::Formation > FormationPtr;

    std::map< std::string, RoleCreator > M_role_factory;
    std::map< std::string, FormationCreator > M_formation_factory;

    FormationPtr M_before_kick_off_formation;
    std::vector< rcsc::Vector2D > M_goal_kick_opp_pos;
    std::vector< rcsc::Vector2D > M_goal_kick_our_pos;
    std::vector< rcsc::Vector2D > M_goalie_catch_opp_pos;
    std::vector< rcsc::Vector2D > M_goalie_catch_our_pos;

    std::vector< PositionType > M_position_types;
    std::vector< rcsc::Vector2D > M_positions;

    FormationPtr M_normal_formation;
    FormationPtr M_offense_formation;
    FormationPtr M_defense_formation;

    FormationPtr M_kickin_our_formation;
    FormationPtr M_setplay_our_formation;

    StrategyType M_opponent_defense_strategy;

public:
    Strategy();

    bool read( const std::string & config_dir );


    rcsc::Vector2D getBeforeKickOffPos( const int number ) const;
    rcsc::Vector2D getPosWhenGoalKickOpp( const int number ) const;
    rcsc::Vector2D getPosWhenGoalKickOur( const int number ) const;
    rcsc::Vector2D getPosWhenGoalieCatchOpp( const int number ) const;
    rcsc::Vector2D getPosWhenGoalieCatchOur( const int number ) const;
    rcsc::Vector2D getSetPlayPosition( const int number,
                                       const rcsc::Formation & formation,
                                       const rcsc::WorldModel & world ) const;

    boost::shared_ptr<
        rcsc::Formation > createFormation( const std::string & type_name ) const;

    boost::shared_ptr<
        SoccerRole > createRole( const int number,
                                 const rcsc::WorldModel & world ) const;

private:
    boost:: shared_ptr<
        const rcsc::Formation > getFormation( const rcsc::WorldModel & world ) const;

    /////////////////////////////////////////////////////////
    bool readStaticPositions( const std::string & filepath,
                              std::vector< rcsc::Vector2D > & positions );

    /////////////////////////////////////////////////////////
    FormationPtr readFormation( const std::string & filepath );

public:

    static
    Strategy & instance()
    {
	static Strategy s_instance;
	return s_instance;
    }

    static
    const
    Strategy & i()
      {
          return instance();
      }

    void updatePosition( const rcsc::WorldModel & wm );

    PositionType getPositionType( const int number ) const
    {
	if ( number < 1 || 11 < number )
	{
		return Position_Center;
	}
	return M_position_types[number - 1];
    }
    rcsc::Vector2D getPosition( const int number ) const
    {
	if ( number < 1 || 11 < number )
	{
		return rcsc::Vector2D::INVALIDATED;
	}

	return M_positions[number - 1];
    }


    StrategyType opponentDefenseStrategy() const
      {
          return M_opponent_defense_strategy;
      }


    enum BallArea {
        BA_CrossBlock, BA_DribbleBlock, BA_DribbleAttack, BA_Cross,
        BA_Stopper,    BA_DefMidField,  BA_OffMidField,   BA_ShootChance,
        BA_Danger,

        BA_None
    };

    static
    BallArea get_ball_area( const rcsc::Vector2D & ball_pos );

    enum P_BallArea {
        PBA_ShootChance1 , PBA_PassShoot   , PBA_Cross , PBA_CrossDribble ,
        PBA_DribbleCross , PBA_DribblePass , PBA_ShootChance2 , PBA_Pass  ,

        PBA_None
    };

    static
    P_BallArea get_penalty_ball_area( const rcsc::Vector2D & ball_pos );

    static
    double get_defense_line_x( const rcsc::WorldModel & wm,
                               const rcsc::Vector2D & home_pos );

    static
    double get_set_play_dash_power( const rcsc::WorldModel & wm );

    static
    double get_defender_dash_power( const rcsc::WorldModel & wm,
                                    const rcsc::Vector2D & home_pos );

};

#endif
